#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta
from __future__ import with_statement

from ionas.fileaccess import (
    MODE_READ,
    MODE_WRITE,
    MODE_APPEND,
    DEFAULT_LOCALE,
)
from ionas.exception import *
from ionas.exception import SizeTooLargeError
from ionas.fileaccess.filelikeobj import FileLikeObj

import stat
import locale

from fxwf.fxwfaccess import FxwfAccess as WorkingFolderAccess
from pkgaddon.fxwf_config import CLIENT_ID
from pkgaddon.fxwf_config import CLIENT_SECRET


def _decode(s, encoding):
    return s if isinstance(s, unicode) else s.decode(encoding)


def to_open_flag(mode):
    if mode == MODE_READ:
        return "r"
    if mode == MODE_WRITE:
        return "w"
    if mode == MODE_APPEND:
        return "w"
    return "r"


class WorkingFolderFileObject(FileLikeObj):

    def __init__(self, file, path, mode, accessor, length, max_size=0):

        FileLikeObj.__init__(self, file, path)
        self.mode = mode
        self.accessor = accessor
        self.length = length
        self._max_size = max_size

    def __len__(self):
        return int(self.accessor.stat(self.path)[stat.ST_SIZE])

    def __read(self, size):
        self.point += size
        return self.file.read(size)

    def __write(self, file_like):
        self.file.write_by_path(file_like)

    def __read_by_path(self, file_like):
        file_like.write_by_path(self.file)

    def __write_by_path(self, file_like):
        self.file.write_by_path(file_like)

    def __close(self):
        self.file.close()

    def read(self, size):
        return self.__read(size)

    def write(self, file_like):
        if self.file is None:
            open_mode = to_open_flag(self.mode)
            try:
                self.length = file_like.get_stat()[stat.ST_SIZE]
            except Exception as err:
                self.length = len(file_like)
            # self.__renew_file()
            self.file = self.accessor.access.open(
                self.path,
                open_mode,
                # length=self.length,
            )
        return self.__write(file_like)

    def write_by_path(self, file_like):
        # TODO
        size_b = file_like.get_stat()[stat.ST_SIZE]
        max_size_b = self.max_size * 1024 * 1024
        if max_size_b and max_size_b < size_b:
            raise SizeTooLargeError
        return self.__write_by_path(file_like)

    def read_by_path(self, file_like):
        return self.__read_by_path(file_like)

    def get_stat(self):
        return self.accessor.stat(self.path)

    def close(self):
        self.__close()

    @property
    def max_size(self):
        return self._max_size


class BackupWorkingFolderAccess(object):

    def __init__(self,
                 target_info,
                 get_proxy=lambda url: (None, None),
                 encoding=""):

        self.sep = "/"

        if encoding:
            self.encoding = encoding
        else:
            self.encoding = locale.getdefaultlocale()[1]
            if not self.encoding:
                self.encoding = DEFAULT_LOCALE

        self.target_info = target_info

        # TODO: REST_BASE_URL from self.fxwf_info.base_url
        from pkgaddon.fxwf_config import REST_BASE_URL
        proxy_host, proxy_port = get_proxy(REST_BASE_URL)
        self._proxy_host = proxy_host
        self._proxy_port = proxy_port

        # self.access = WorkingFolderAccess(
        #     target_info.email,
        #     target_info.password,
        #     CLIENT_ID,
        #     CLIENT_SECRET,
        #     target_info.cabinet_id,
        #     target_info.drawer_name,
        #     encoding=self.encoding,
        #     proxy_host=proxy_host,
        #     proxy_port=proxy_port,
        # )

    @property
    def access(self):
        access = getattr(self, '_access', None)
        if access is None:
            access = WorkingFolderAccess(
                self.target_info.email,
                self.target_info.password,
                CLIENT_ID,
                CLIENT_SECRET,
                self.target_info.cabinet_id,
                self.target_info.drawer_name,
                encoding=self.encoding,
                proxy_host=self._proxy_host,
                proxy_port=self._proxy_port,
            )
            setattr(self, '_access', access)
        return access

    def listdir(self, path):
        _path = _decode(path, self.encoding)
        _path = self.__pathnormalize(_path)
        return [e.encode(self.encoding) for e in self.access.listdir(_path)]

    def isdir(self, entry):
        # FIXME
        raise Exception('isdir')

    def open(self, path, mode):
        _path = path
        # _path = _decode(path, self.encoding)
        _path = self.__pathnormalize(_path)
        open_mode = to_open_flag(mode)
        # opened_file = None
        # if open_mode == "r":
        #     opened_file = self.access.open(_path, open_mode)
        opened_file = self.access.open(_path, open_mode)
        return WorkingFolderFileObject(
            opened_file,
            _path,
            mode,
            accessor=self,
            length=-1,
            max_size=self.target_info.max_upload_size,
        )

    def stat(self, path):
        _path = path
        # _path = _decode(_path, self.encoding)
        _path = self.__pathnormalize(_path)
        # TODO
        from fxwf.fxwfinfo import ObjectNotFoundHttpError
        try:
            return self.access.stat(_path)
        except ObjectNotFoundHttpError as err:
            raise NoEntryError(err, _path)

    def chmod(self, path, mode):
        return NotImplementedError

    def utime(self, path, mtime):
        return NotImplementedError

    def remove(self, path):
        _path = path
        # _path = _decode(_path, self.encoding)
        _path = self.__pathnormalize(_path)
        self.access.remove(_path)

    def rmdir(self, path):
        _path = path
        _path = _decode(_path, self.encoding)
        _path = self.__pathnormalize(_path)
        # TODO
        from fxwf.fxwfinfo import ObjectNotFoundHttpError
        try:
            self.access.file_client.delete_directory(_path)
        except ObjectNotFoundHttpError as err:
            raise NoEntryError(err, _path)

    def mkdir(self, path, p=True):
        # _path = _decode(path, self.encoding)
        path = self.__pathnormalize(path)
        if path in ('', self.sep):
            pass
        else:
            # TODO
            from fxwf.fxwfinfo import ObjectNotFoundHttpError
            try:
                self.access.mkdir(path)
            except ObjectNotFoundHttpError as err:
                if p:
                    _head, _tail = self.__pathsplit(path)
                    self.mkdir(_head, p=p)
                    self.mkdir(path, p=False)
                else:
                    raise
                    # raise NoEntryError(err, path)
            except Exception as err:
                # TODO: alreadyExists
                if self.access.isdir(path):
                    pass
                else:
                    raise

    def move(self, src, dst):
        _src_path = _decode(src, self.encoding)
        _src_path = self.__pathnormalize(_src_path)
        _dst_path = _decode(dst, self.encoding)
        _dst_path = self.__pathnormalize(_dst_path)
        # TODO
        from fxwf.fxwfinfo import ObjectNotFoundHttpError
        _src_head, _src_tail = self.__pathsplit(_src_path)
        _dst_head, _dst_tail = self.__pathsplit(_dst_path)
        try:
            self.access.file_client.rename_directory(_src_path, _dst_tail)
            _src_path2 = self.__pathjoin(_src_head, _dst_tail)
            self.access.file_client.move_directory(_src_path2, _dst_head)
        except ObjectNotFoundHttpError as err:
            raise NoEntryError(err, src)

    def __pathsplit(self, path):
        separator = self.sep
        names = path.split(separator)
        head = separator.join(names[:-1])
        tail = names[-1]
        return head, tail

    def __pathjoin(self, head, tail):
        separator = self.sep
        return separator.join([head, tail])

    def __pathnormalize(self, path):
        # TODO
        path = path.replace('{0}{0}'.format(self.sep), '{0}'.format(self.sep))
        path = path.rstrip(self.sep)
        return path
